import numpy as np
from innovations import Innovations
import models

class Modern():
    def __init__(self, mode, plot=True):
        self.log = []
        if mode == 'ukf1':
            m = models.Jazwinski1()
            self.ukf1(m)
        elif mode == 'ukf2':
            m = models.Jazwinski2()
            self.ukf2(m)
        innov = Innovations(self.log)
        if plot: innov.plot_standard()

    def ukf1(self, m):
        xhat, Ptil = 2.2, .01
        W = np.array([m.k1, m.k2, m.k2])
        for step in m.steps():
            X = m.va(m.X(xhat, Ptil))
            Y = m.vc(m.Xhat(X, m.Rww))
            Rksiksi = W @ np.power(Y - W @ Y, 2) + m.Rvv
            RXtilksi = W @ np.multiply(X - W @ X, Y - W @ Y)
            K = RXtilksi / Rksiksi
            yhat = W @ Y
            xhat = W @ X + K * (step[2] - W @ Y)
            Ptil = W @ np.power(X - W @ X, 2) + m.Rww - K * Rksiksi * K
            self.log.append([step[0], xhat, yhat, step[1] - xhat, step[2] - yhat])

    def ukf2(self, m):
        xhat = np.array([2.0, .055, .044])
        Ptil = .1 * np.eye(3)
        W = np.array([m.k1, m.k2, m.k2, m.k2, m.k2, m.k2, m.k2])
        for step in m.steps():
            X = m.va(m.X(xhat, Ptil))
            Y = m.vc(m.Xhat(X, m.Rww))
            Rksiksi = W * m.ksi(Y, W) @ m.ksi(Y, W).T + m.Rvv
            RXtilksi = W * m.Xtil(X, W) @ m.ksi(Y, W).T
            K = np.squeeze(np.asarray(RXtilksi / Rksiksi))
            yhat = W @ Y
            xhat = W @ X.T + K * (step[2] - W @ Y)
            Ptil = Ptil # W * m.Xtil(X, W) @ m.Xtil(X, W).T + m.Rww - K @ K.T * Rksiksi
            self.log.append([step[0], xhat[0], yhat, step[1][0] - xhat[0], step[2] - yhat])

if __name__ == "__main__":
    Modern('ukf1')
    Modern('ukf2')




import math
import numpy as np

class Jazwinski1():
    def __init__(self):
        self.tsteps = 151
        self.dt = .01
        self.x = 2.
        self.Rww = 1e-6
        self.Rvv = 9e-2
        self.log = []
        self.va = np.vectorize(self.a)
        self.vc = np.vectorize(self.c)
        self.bignsubx = 1
        self.kappa = 1
        self.k0 = self.bignsubx + self.kappa
        self.k1 = self.kappa / float(self.k0)
        self.k2 = 1 / float(2 * self.k0)
        # self.vAcurl = np.vectorize(self.Acurl)
        # self.vCcurl = np.vectorize(self.Ccurl)
        self.nsamp = 250

    def a(self, x, w=0):
        return (1 - .05 * self.dt) * x + (.04 * self.dt) * x ** 2 + w

    def c(self, x, v=0):
        return x ** 2 + x ** 3 + v

    def A(self, x):
        return 1 - .05 * self.dt + .08 * self.dt * x

    def C(self, x):
        return 2 * x + 3 * x ** 2

    def X(self, xhat, Ptil):
        return [xhat, xhat + math.sqrt(self.k0 * Ptil), xhat - math.sqrt(self.k0 * Ptil)]

    def Xhat(self, X, Rww):
        return [X[0], X[1] + self.kappa * math.sqrt(Rww), X[2] - self.kappa * math.sqrt(Rww)]

    def Apf(self, x):
        return (1 - .05 * self.dt) * x + (.04 * self.dt) * x ** 2 + math.sqrt(self.Rww) * np.random.randn(self.nsamp)

    def Cpf(self, y, x):
        return np.exp(-np.log(2. * np.pi * self.Rvv) / 2. - (y - x**2 - x**3)**2 / (2. * self.Rvv))

    def steps(self):
        for tstep in range(self.tsteps):
            tsec = tstep * self.dt
            w = math.sqrt(self.Rww) * np.random.randn()
            v = math.sqrt(self.Rvv) * np.random.randn()
            self.x = self.a(self.x, w)
            self.y = self.c(self.x, v)
            if tstep == 0: continue
            self.log.append([tsec, self.x, self.y])
            yield (tsec, self.x, self.y)

class Jazwinski2():
    def __init__(self):
        self.tsteps = 1501
        self.dt = .01
        self.x = np.array([2., .05, .04])
        self.Rww = 1e-9 * np.array([1, 1, 1])
        self.Rvv = 9e-2
        self.log = []
        self.n = 3
        self.kappa = 1
        self.alpha = 1
        self.beta = 2
        self.nk = self.n + self.kappa
        self.lam = self.alpha**2 * self.nk - self.n
        self.nl = self.n + self.lam
        self.k1 = self.lam / float(self.nl)
        self.k2 = 1 / float(2 * self.nl)
        self.W0y = self.lam / float(self.nl)
        self.W0 = self.lam / float(self.nl) + (1 - self.alpha**2 + self.beta)
        self.nsamp = 250

    def a(self, x, w=0):
        return np.array([(1 - x[1] * self.dt) * x[0] + x[2] * self.dt * x[0] ** 2, x[1], x[2]]) + w

    def c(self, x, v=0):
        return x[0] ** 2 + x[0] ** 3 + v

    def A(self, x):
        A = np.eye(3)
        A[0, 0] = 1 - x[1] * self.dt + 2 * x[2] * self.dt * x[0]
        A[0, 1] = -self.dt * x[0]
        A[0, 2] = self.dt * x[0] ** 2
        return A

    def C(self, x):
        return np.array([2 * x[0] + 3 * x[0] ** 2, 0, 0])

    def X(self, x, P):
        C = np.linalg.cholesky(P)
        X = np.zeros([3, 7])
        X[:, 0] = x
        X[:, 1] = x + math.sqrt(self.nl) * C.T[:, 0]
        X[:, 2] = x + math.sqrt(self.nl) * C.T[:, 1]
        X[:, 3] = x + math.sqrt(self.nl) * C.T[:, 2]
        X[:, 4] = x - math.sqrt(self.nl) * C.T[:, 0]
        X[:, 5] = x - math.sqrt(self.nl) * C.T[:, 1]
        X[:, 6] = x - math.sqrt(self.nl) * C.T[:, 2]
        return X

    def Xhat(self, X, Rww):
        Xhat = np.zeros([3, 7])
        Xhat[:, 0] = X[:, 0]
        Xhat[:, 1] = X[:, 1] + np.array([self.kappa * math.sqrt(Rww[0]), 0, 0])
        Xhat[:, 2] = X[:, 2] + np.array([0, self.kappa * math.sqrt(Rww[1]), 0])
        Xhat[:, 3] = X[:, 3] + np.array([0, 0, self.kappa * math.sqrt(Rww[2])])
        Xhat[:, 4] = X[:, 4] - np.array([self.kappa * math.sqrt(Rww[0]), 0, 0])
        Xhat[:, 5] = X[:, 5] - np.array([0, self.kappa * math.sqrt(Rww[1]), 0])
        Xhat[:, 6] = X[:, 6] - np.array([0, 0, self.kappa * math.sqrt(Rww[2])])
        return Xhat
        # C = np.linalg.cholesky(Rww)
        # Xhat[:, 1] = X[:, 1] + self.nl * C.T[:, 0]
        # Xhat[:, 2] = X[:, 2] + self.nl * C.T[:, 1]
        # Xhat[:, 3] = X[:, 3] + self.nl * C.T[:, 2]
        # Xhat[:, 4] = X[:, 4] + self.nl * C.T[:, 0]
        # Xhat[:, 5] = X[:, 5] + self.nl * C.T[:, 1]
        # Xhat[:, 6] = X[:, 6] + self.nl * C.T[:, 2]

    def va(self, X):
        for i in range(7): X[:, i] = self.a(X[:, i], 0)
        return X

    def vc(self, Xhat):
        Y = np.zeros(7)
        for i in range(7): Y[i] = self.c(Xhat[:, i], 0)
        return Y

    def Xtil(self, X, W):
        Xtil = np.zeros((3, 7))
        for i in range(7): Xtil[:, i] = X[:, i] - W @ X.T
        return Xtil

    def ksi(self, Y, W):
        ksi = np.zeros((1, 7))
        for i in range(7): ksi[0, i] = Y[i] - W @ Y.T
        return ksi

    def Apf(self, x):
        return (1 - x[1] * self.dt) * x[0] + x[2] * self.dt * x[0] ** 2

    def Cpf(self, y, x):
        return np.exp(-np.log(2. * np.pi * self.Rvv) / 2. - (y - x**2 - x**3)**2 / (2. * self.Rvv))

    def steps(self):
        for tstep in range(self.tsteps):
            tsec = tstep * self.dt
            w = np.multiply(np.random.randn(1, 3), np.sqrt(np.diag(self.Rww)))
            v = math.sqrt(self.Rvv) * np.random.randn()
            self.x = self.a(self.x, w[0])
            self.y = self.c(self.x, v)
            if tstep == 0: continue
            self.log.append([tsec, self.x, self.y])
            yield (tsec, self.x, self.y)

if __name__ == "__main__":
    sim = Jazwinski1()
    for step in sim.steps():
        print(step)




import argparse

if __name__ == "__main__":
    parser = argparse.ArgumentParser('statespace')
    parser.add_argument('-d', '--demo', help='demo', action='store_true')
    args = parser.parse_args()
    if args.demo:
        from classical import Classical
        Classical('ekf2')



import numpy as np
import matplotlib.pyplot as plt
import math

class Innovations():
    def __init__(self, log):
        self.log = np.asarray(log)
        self.tts = self.log[:,0]
        self.ets = self.log[:,4]
        self.zmw()
        self.lw = 1

    def autocorr1(self, x):
        n = x.size
        ac = np.zeros([n, 1])
        for k in range(1, n): # k = 1 to n-1
            ac[k] = 0
            for t in range(1, n-k): # t = 1 to n-k
                ac[k] += (self.ets[t-1] - self.mhate) * (self.ets[t-1+k] - self.mhate) #/ (n-k)
        return ac

    def autocorr2(self, x):
        ac = np.correlate(x, x, mode='full')
        return ac[ac.size//2:]

    def zmw(self):
        self.mhate = np.mean(self.ets)
        self.Rhatee = np.mean(np.power(self.ets, 2))
        n = len(self.ets)
        self.tau = 1.96 * math.sqrt(self.Rhatee / n)
        self.iszeromean = True
        if abs(self.mhate) > self.tau:
            self.iszeromean = False
        return self.autocorr1(self.ets)

    def wssr(self):
        return

    def plot_standard(self):
        x1 = 3
        plt.subplot(3, 2, 1), plt.plot(self.tts[x1:], self.log[x1:,1], linewidth=self.lw), plt.ylabel('xhat')
        plt.subplot(3, 2, 2), plt.plot(self.tts[x1:], self.log[x1:,3], linewidth=self.lw), plt.ylabel('err')
        plt.subplot(3, 2, 3), plt.plot(self.tts[x1:], self.log[x1:,2], linewidth=self.lw), plt.ylabel('yhat')
        plt.subplot(3, 2, 4), plt.plot(self.tts[x1:], self.log[x1:,4], linewidth=self.lw), plt.ylabel('innov')
        plt.subplot(3, 2, 5), plt.plot(self.tts[x1:], self.autocorr1(self.ets)[x1:], linewidth=self.lw), plt.ylabel('autocorr')
        plt.show()

    def abp(self, tts, xhatts, xtilts, yhatts):
        def tsplot(x, y):
            plt.plot(x, y, marker='o', markersize=1, linewidth=0)
        plt.subplot(3, 2, 1)
        tsplot(tts[1:], xhatts[1:, 0])
        plt.subplot(3, 2, 2)
        tsplot(tts[1:], xhatts[1:, 1])
        plt.subplot(3, 2, 3)
        tsplot(tts[1:], xhatts[1:, 2])
        plt.subplot(3, 2, 4)
        tsplot(tts[1:], self.ets[1:])
        plt.subplot(3, 2, 5)
        tsplot(tts[1:], yhatts[1:])
        plt.subplot(3, 2, 6)
        tsplot(tts[1:], self.Rhatee[1:])
        plt.show()



import numpy as np
from innovations import Innovations
import models

def ud_factorization(M):
    assert np.allclose(M, M.T)
    n = M.shape[0]
    M = np.triu(M)
    U = np.eye(n)
    d = np.zeros(n)
    for j in reversed(range(2, n + 1)):
        d[j - 1] = M[j - 1, j - 1]
        if d[j - 1] > 0:
            alpha = 1.0 / d[j - 1]
        else:
            alpha = 0.0
        for k in range(1, j):
            beta = M[k - 1, j - 1]
            U[k - 1, j - 1] = alpha * beta
            M[0:k, k - 1] = M[0:k, k - 1] - beta * U[0:k, j - 1]
    d[0] = M[0, 0]
    return U, np.diag(d)

def thornton_temporal_update(xin, Phi, Uin, Din, Gin, Q):
    x, U, D = Phi @ xin, Uin, Din
    n, r = 3, 3
    G = Gin
    U = np.eye(3)
    PhiU = Phi @ Uin
    for i in reversed(range(3)):
        sigma = 0
        for j in range(n):
            sigma = sigma + PhiU[i, j] ** 2 * Din[j, j]
            if (j <= r - 1):
                sigma = sigma + G[i, j] ** 2 + Q[j, j]
        D[i, i] = sigma
        ilim = i - 1
        if not ilim < 0:
            for j in range(ilim):
                sigma = 0
                for k in range(n):
                    sigma = sigma + PhiU[i, k] * Din[k, k] * PhiU[j, k]
                for k in range(r):
                    sigma = sigma + G[i, k] * Q[k, k] * G[j, k]
                U[j, i] = sigma / D[i, i]
                for k in range(n):
                    PhiU[j, k] = PhiU[j, k] - U[j, i] * PhiU[i, k]
                for k in range(r):
                    G[j, k] = G[j, k] - U[j, i] * G[i, k]
    return x, U, D

def bierman_observational_update(z, R, H, xin, Uin, Din):
    x, U, D = xin, Uin, Din
    a = U.T @ H.T
    b = D @ a
    dz = z  # z - H @ xin
    alpha = R
    gamma = 1 / alpha
    for j in range(3):
        beta = alpha
        alpha = alpha + a[j] * b[j]
        lamda = -a[j] * gamma
        gamma = 1 / alpha
        D[j, j] = beta * gamma * D[j, j]
        jlim = j - 1
        if not jlim < 0:
            for i in range(jlim):
                beta = U[i, j]
                U[i, j] = beta + b[i] * lamda
                b[i] = b[i] + b[j] * beta
    dzs = gamma * dz
    x = x + dzs * b
    return x, U, D

class Classical():
    def __init__(self, mode, plot=True):
        self.log = []
        if mode == 'ekf1':
            m = models.Jazwinski1()
            self.ekf1(m)
        elif mode == 'ekf2':
            m = models.Jazwinski2()
            self.ekf2(m)
        innov = Innovations(self.log)
        if plot: innov.plot_standard()

    def ekf1(self, m):
        xhat = 2.2
        Ptil = .01
        for step in m.steps():
            xhat = m.a(xhat)
            Ptil = m.A(xhat) * Ptil * m.A(xhat)
            Ree = m.C(xhat) * Ptil * m.C(xhat) + m.Rvv
            K = Ptil * m.C(xhat) / Ree
            yhat = m.c(xhat)
            xhat = xhat + K * (step[2] - yhat)
            Ptil = (1 - K * m.C(xhat)) * Ptil
            self.log.append([step[0], xhat, yhat, step[1]-xhat, step[2]-yhat])

    def ekf2(self, m):
        xhat = np.array([2, .055, .044])
        Ptil = 1. * np.eye(3)
        U, D = ud_factorization(Ptil)
        for step in m.steps():
            xhat, U, D = thornton_temporal_update(xin=m.a(xhat), Phi=m.A(xhat), Uin=U, Din=D, Gin=np.eye(3), Q=np.diag(m.Rww))
            yhat = m.c(xhat)
            xhat, U, D = bierman_observational_update(z=step[2] - yhat, R=m.Rvv, H=m.C(xhat), xin=xhat, Uin=U, Din=D)
            self.log.append([step[0], xhat[0], yhat, step[1][0]-xhat[0], step[2]-yhat])

if __name__ == "__main__":
    Classical('ekf1')
    Classical('ekf2')



# monte carlo sampling processor, bootstrap particle filter
import numpy as np
import math
from innovations import Innovations
import models

def resample(xi, Wi):
    tmp = []
    for i in range(xi.size):
        tmp.append([xi[i], Wi[i]])
    tmp = sorted(tmp, key=lambda x: x[0])
    cdf = [[tmp[0][0], tmp[0][1]]]
    for i in range(1, len(tmp)):
        cdf.append([tmp[i][0], tmp[i][1] + cdf[i - 1][1]])
    cdf = np.asarray(cdf)
    uk = np.sort(np.random.uniform(size=xi.size))
    xhati, k = [], 0
    for row in cdf:
        while k < uk.size and uk[k] <= row[1]:
            xhati.append(row[0])
            k += 1
    xhati = np.asarray(xhati)
    assert xhati.size == xi.size
    return xhati

def roughen(x):
    return x + .1 * np.random.randn(x.size)

def normalize(W):
    return W / sum(W)

class Particle():
    def __init__(self, mode, plot=True):
        self.log = []
        if mode == 'pf1':
            m = models.Jazwinski1()
            self.pf1(m)
        if mode == 'pf2':
            m = models.Jazwinski2()
            self.pf2(m)
        innov = Innovations(self.log)
        if plot: innov.plot_standard()

    def pf1(self, m):
        xhat = 2.05
        x = xhat + math.sqrt(m.Rww) * np.random.randn(m.nsamp)
        W = normalize(np.ones(m.nsamp))
        for step in m.steps():
            x = resample(x, W)
            x = m.Apf(x)
            W = normalize(m.Cpf(step[2], x))
            yhat = m.c(xhat, 0)
            xhat = W @ x
            self.log.append([step[0], xhat, yhat, step[1] - xhat, step[2] - yhat])

    def pf2(self, m):
        xhat = np.array([2.0, .055, .044])
        x = xhat + np.sqrt(m.Rww) * np.random.randn(m.nsamp, 3)
        W = np.ones((m.nsamp, 3)) / m.nsamp
        for step in m.steps():
            x[:, 0], x[:, 1], x[:, 2] = resample(x[:, 0], W[:, 0]), roughen(x[:, 1]), roughen(x[:, 2])
            x[:, 0] = np.apply_along_axis(m.Apf, 1, x)
            W[:, 0] = normalize(m.Cpf(step[2], x[:, 0]))
            yhat = m.c(xhat, 0)
            xhat = [W[:, 0].T @ x[:, 0], W[:, 1].T @ x[:, 1], W[:, 2].T @ x[:, 2]]
            self.log.append([step[0], xhat[0], yhat, step[1][0] - xhat[0], step[2] - yhat])

if __name__ == "__main__":
    Particle('pf1')
    Particle('pf2')
